﻿using System;
using System.Security.Authentication;
using System.Text;
using System.Threading;
using Unity.WebRTC;
using UnityEngine;
using WebSocketSharp;

namespace PureWeb
{
    /// <summary>
    /// Communicates with the PureWeb Platform using end-of-life websocket protocol.
    /// Requires app-server 6.x.
    /// </summary>
    public class Signaling 
    {
        private string m_url;
        private float m_timeout;
        private bool m_running;
        private Thread m_signalingThread;
        private AutoResetEvent m_wsCloseEvent;
        private WebSocket m_webSocket;

        public Signaling(string url, float timeout)
        {
            m_url = url;
            m_timeout = timeout;
            m_wsCloseEvent = new AutoResetEvent(false);
        }

        public void Start()
        {
            m_running = true;
            m_signalingThread = new Thread(WSManage);
            m_signalingThread.IsBackground = true;
            m_signalingThread.Start();
        }

        public void Close()
        {
            m_webSocket?.Close();
        }


        public void Stop()
        {
            m_running = false;
            m_webSocket?.Close();
        }

        public delegate void OnOfferHandler(Signaling signalling, PlatformSessionDescription offer);
        public event OnOfferHandler OnOffer;

        public delegate void OnConfigHandler(Signaling signalling, PlatformClientConfiguration configuration);
        public event OnConfigHandler OnConfig;

        public delegate void OnIceCandidateHandler(Signaling signalling, PlatformIceCandidateDescription candidate);
        public event OnIceCandidateHandler OnIceCandidate;

        public void SendOffer()
        {
            throw new NotImplementedException();
        }

        public void SendAnswer(int playerId, RTCSessionDescription answer)
        {
            Debug.Log("Sending answer");
            PlatformSessionDescription data = new PlatformSessionDescription();
            data.playerId = playerId;
            data.sdp = answer.sdp;
            data.type = "answer";

            WSSend(data);
        }

        public void SendCandidate(int playerId, RTCIceCandidate candidate)
        {
            PlatformIceCandidateDescription data = new PlatformIceCandidateDescription();
            data.playerId = playerId;
            data.type = "iceCandidate";
            data.candidate = new PlatformIceCandidate
            {
                candidate = candidate.Candidate,
                sdpMLineIndex = candidate.SdpMLineIndex ?? 0,
                sdpMid = candidate.SdpMid
            };
            WSSend(data);
        }

        private void WSManage()
        {
            while (m_running)
            {
                WSCreate();

                m_wsCloseEvent.WaitOne();

                Thread.Sleep((int)(m_timeout * 1000));
            }

            Debug.Log("Signaling: WS managing thread ended");
        }

        private void WSCreate()
        {
            m_webSocket = new WebSocket(m_url);
            if (m_url.StartsWith("wss"))
            {
                m_webSocket.SslConfiguration.EnabledSslProtocols =
                    SslProtocols.Tls | SslProtocols.Tls11 | SslProtocols.Tls12;
            }

            m_webSocket.OnOpen += WSConnected;
            m_webSocket.OnMessage += WSProcessMessage;
            m_webSocket.OnError += WSError;
            m_webSocket.OnClose += WSClosed;

            Monitor.Enter(m_webSocket);

            Debug.Log($"Signaling: Connecting WS {m_url}");
            m_webSocket.ConnectAsync();
        }

        private void WSProcessMessage(object sender, MessageEventArgs e)
        {
            var content = Encoding.UTF8.GetString(e.RawData);
            Debug.Log($"Signaling: Receiving message: {content}");

            try
            {
                var cirrusMessage = JsonUtility.FromJson<PlatformMessage>(content);
                if ("config" == cirrusMessage.type)
                {
                    Debug.Log("Signaling: Received config message");
                    var config = JsonUtility.FromJson<PlatformClientConfiguration>(content);
                    OnConfig?.Invoke(this, config);
                }
                else if ("offer" == cirrusMessage.type)
                {
                    var offer = JsonUtility.FromJson<PlatformSessionDescription>(content);
                    if (!string.IsNullOrEmpty(offer.sdp))
                    {
                        OnOffer?.Invoke(this, offer);
                    }
                    else
                    {
                        Debug.LogError("Signaling: Received message from unknown peer");
                    }
                }
                else if ("iceCandidate" == cirrusMessage.type)
                {
                    var candidate = JsonUtility.FromJson<PlatformIceCandidateDescription>(content);
                    OnIceCandidate?.Invoke(this, candidate);
                }
                else
                {
                    Debug.LogError("Signaling: Cannot handle message of provided type");
                }
            }
            catch (Exception ex)
            {
                Debug.LogError("Signaling: Failed to parse message: " + ex);
            }
        }

        private void WSConnected(object sender, EventArgs e)
        {
            Debug.Log("Signaling: WS connected.");
        }


        private void WSError(object sender, ErrorEventArgs e)
        {
            Debug.LogError($"Signaling: WS connection error: {e.Message}");
        }

        private void WSClosed(object sender, CloseEventArgs e)
        {
            Debug.LogError($"Signaling: WS connection closed, code: {e.Code}");

            m_wsCloseEvent.Set();
            m_webSocket = null;
        }

        private void WSSend(object data)
        {
            if (m_webSocket == null || m_webSocket.ReadyState != WebSocketState.Open)
            {
                Debug.LogError("Signaling: WS is not connected. Unable to send message");
                return;
            }

            if (data is string s)
            {
                Debug.Log("Signaling: Sending WS data: " + s);
                m_webSocket.Send(s);
            }
            else
            {
                string str = JsonUtility.ToJson(data);
                Debug.Log("Signaling: Sending WS data: " + str);
                m_webSocket.Send(str);
            }
        }
    }
}
