﻿using System.Collections.Generic;
using UnityEngine.InputSystem.EnhancedTouch;
using Unity.WebRTC;
using UnityEngine;
using System.Text.RegularExpressions;
using System;
using System.Collections;
using System.Reflection;
using System.Text;
using StreamingAgentPlugin;
using Unity.Collections;

namespace PureWeb
{
    using DataChannelDictionary = Dictionary<int, RTCDataChannel>;

    public class InteractionEventArgs : EventArgs
    {
        public int PlayerId;
        public int Size;
        public byte[] Bytes;

        public InteractionEventArgs(int playerId, byte[] data)
        {
            this.PlayerId = playerId;

            if (data != null && data.Length >= 3)
            {
                Bytes = new byte[data.Length - 3];

                byte[] messageLength = new byte[2];

                Array.Copy(data, 1, messageLength, 0, 2);
                Array.Copy(data, 3, Bytes, 0, Bytes.Length);

                Size = BitConverter.ToUInt16(messageLength, 0);
            }
            else
            {
                Debug.LogError("Unexpected message data format");
                this.Bytes = data;
            }

        }
    }

    public class PlayerConnectedEventArgs : EventArgs
    {
        public int PlayerId;
    }

    [RequireComponent(typeof(AudioSource))]
    public class PureWebStreaming : MonoBehaviour
    {
        [SerializeField, Tooltip("Target camera texture (defaults to screen). Camera must be tagged as PureWebCamera")]
        private bool m_targetCamera = false;

        [SerializeField, Tooltip("Streaming width")]
        private int m_width = 1920;

        [SerializeField, Tooltip("Streaming height")]
        private int m_height = 1080;

        private Signaling m_signaling;
        private readonly Dictionary<int, RTCPeerConnection> m_mapConnectionIdAndPeer = new Dictionary<int, RTCPeerConnection>();
        private readonly Dictionary<RTCPeerConnection, DataChannelDictionary> m_mapPeerAndChannelDictionary = new Dictionary<RTCPeerConnection, DataChannelDictionary>();
        private readonly Dictionary<RTCDataChannel, PureWebInputAdapter> m_mapChannelAndRemoteInput = new Dictionary<RTCDataChannel, PureWebInputAdapter>();
        private readonly List<PureWebCameraController> m_listController = new List<PureWebCameraController>();
        private readonly Dictionary<VideoStreamTrack, Action> m_listVideoStreamTrackUpdaters = new Dictionary<VideoStreamTrack, Action>();
        private readonly Dictionary<RTCPeerConnection, RTCRtpSender> m_rtpVideoSenders = new Dictionary<RTCPeerConnection, RTCRtpSender>();
        private MediaStream m_audioStream;
        private PlatformClientConfiguration m_clientConfiguration;

        private Texture2D m_screen;
        private Configuration m_configuration;

        public static PureWebStreaming Instance { get; private set; }

        public SidecarClient SidecarClient { get; private set; }

        public event EventHandler<InteractionEventArgs> InteractionEvent;
        public event EventHandler<PlayerConnectedEventArgs> PlayerConnectedEvent;


        public void Awake()
        {
            m_configuration = new Configuration();

            DontDestroyOnLoad(this.gameObject);
            Instance = this;
            WebRTC.Initialize(EncoderType.Hardware);
            EnhancedTouchSupport.Enable();

            if (m_targetCamera)
            {
                Debug.LogError("Camera capture not currently supported.");
            }
        }
        private void CreateOrUpdateTrack()
        {
            if (m_screen == null)
            {
                Screen.SetResolution(m_width, m_height, false, 60);
                m_screen = new Texture2D(Screen.width, Screen.height, TextureFormat.BGRA32, false);
            }
            else
            {
                m_screen.Resize(Screen.width, Screen.height);
            }

            var videoTrack = new VideoStreamTrack("screen", m_screen);
            // WebRTC.Update begins an infinite loop and does not allow us to perform our own capture before calling Update on the individual tracks.
            // Additionally, Track.Update is internal. This should be addressed.
            var trackUpdateMethod = videoTrack.GetType().GetMethod("Update", BindingFlags.NonPublic | BindingFlags.Instance);
            var update = (Action)trackUpdateMethod.CreateDelegate(typeof(Action), videoTrack);
            m_listVideoStreamTrackUpdaters.Add(videoTrack, update);
        }

        public void OnDestroy()
        {
            Instance = null;
            EnhancedTouchSupport.Disable();
            WebRTC.Dispose();
            PureWebInputAdapterFactory.Dispose();
            Unity.WebRTC.Audio.Stop();
        }
        public void Start()
        {
            if (AudioSettings.speakerMode != AudioSpeakerMode.Stereo)
            {
                Debug.LogError("Only stereo audio capture is supported.");
            }
            else
            {
                AudioRenderer.Start();
                m_audioStream = Unity.WebRTC.Audio.CaptureStream();
            }
            
            if (SidecarClient == null && m_configuration.IsSidecarEnabled)
            {
                SidecarClient = new SidecarClient(m_configuration.SidecarUrl);
            }
        }

        public void Update()
        {
            StartCoroutine(CaptureScreen());
        }

        private IEnumerator CaptureScreen()
        {


            // Are we shutting down?
            if (Instance == null) yield return null;

            if (m_screen == null || (Screen.width != m_screen.width || Screen.height != m_screen.height))
            {
                CreateOrUpdateTrack();
            }


            // Capture audio frames
            if (m_audioStream != null)
            {
                try
                {
                    var sampleCountFrame = AudioRenderer.GetSampleCountForCaptureFrame();
                    var length = sampleCountFrame * 2; // Channel count is checked in Start()
                    var buffer = new NativeArray<float>(length, Allocator.Temp);

                    AudioRenderer.Render(buffer);
                    Audio.Update(buffer.ToArray(), buffer.Length);
                }
                catch (Exception e)
                {
                    Debug.LogException(e);
                }
            }
            

            yield return new WaitForEndOfFrame();
            m_screen.ReadPixels(new Rect(0, 0, Screen.width, Screen.height), 0, 0);
            m_screen.Apply();

            foreach (var update in m_listVideoStreamTrackUpdaters.Values)
            {
                try
                {
                    update();
                }
                catch (Exception e)
                {
                    Debug.LogErrorFormat("Failure during track update {0}", e);
                }
            }

        }


        void OnEnable()
        {
            if (m_signaling == null)
            {
                m_signaling = new Signaling(m_configuration.StreamerUrl, 5);

                m_signaling.OnConfig += OnConfig;
                m_signaling.OnOffer += OnOffer;
                m_signaling.OnIceCandidate += OnIceCandidate;
            }
            m_signaling.Start();
        }

        void OnDisable()
        {
            if (m_signaling != null)
            {
                m_signaling.Stop();
                m_signaling = null;
            }

            foreach (var peerDataChannels in m_mapPeerAndChannelDictionary)
            {
                var dataChannelMap = peerDataChannels.Value;
                foreach (var dataChannelId in dataChannelMap)
                {
                    var dc = dataChannelId.Value;
                    dc.Close();

                    // Our callback may not be called because we're shutting down, manually clean-up here.
                    OnCloseChannel(dc);
                }
                dataChannelMap.Clear();

                peerDataChannels.Key.Close(); // PeerConnection
            }

            m_mapConnectionIdAndPeer.Clear();
        }

        void OnConfig(Signaling signaling, PlatformClientConfiguration clientConfiguration)
        {
            m_clientConfiguration = clientConfiguration;
        }

        void OnOffer(Signaling signaling, PlatformSessionDescription e)
        {
            BeginNegotiation(signaling, e);
        }

        void BeginNegotiation(Signaling signaling, PlatformSessionDescription e)
        {
            RTCSessionDescription offer;
            offer.type = RTCSdpType.Offer;
            offer.sdp = e.sdp;
            var connectionId = e.playerId;
            if (m_mapConnectionIdAndPeer.TryGetValue(connectionId, out RTCPeerConnection target))
            {
                Debug.Log("Removing existing player: " + connectionId);
                target.Close();
                m_mapConnectionIdAndPeer.Remove(connectionId);
            }

            var pc = new RTCPeerConnection();
            if (!m_mapConnectionIdAndPeer.ContainsKey(e.playerId))
            {
                m_mapConnectionIdAndPeer.Add(e.playerId, pc);
            }

            pc.OnDataChannel = new DelegateOnDataChannel(channel => { OnDataChannel(pc, e.playerId, channel); });
            pc.OnIceCandidate = new DelegateOnIceCandidate(candidate =>
            {
                signaling.SendCandidate(e.playerId, candidate);
            });
            pc.OnIceConnectionChange = new DelegateOnIceConnectionChange(state =>
            {
                Debug.Log("PeerConnection ICE Connection State:" + state);
                if (state == RTCIceConnectionState.Disconnected)
                {
                    pc.Close();
                    m_mapConnectionIdAndPeer.Remove(e.playerId);
                    if (e.playerId == 1)
                    {
                        signaling.Close();
                    }
                }
            });

            var conf = new RTCConfiguration
            {
                iceServers = m_clientConfiguration.peerConnectionOptions.iceServers,
                iceCandidatePoolSize = m_clientConfiguration.peerConnectionOptions.iceCandidatePoolSize == 0 ? 4 : m_clientConfiguration.peerConnectionOptions.iceCandidatePoolSize,
                iceTransportPolicy = m_clientConfiguration.peerConnectionOptions.iceTransportPolicy == "relay" ? RTCIceTransportPolicy.Relay : RTCIceTransportPolicy.All
            };

            Debug.Log("SetConfiguration: " + JsonUtility.ToJson(conf));

            var configError = pc.SetConfiguration(ref conf);
            Debug.Log("Config error: " + configError);
            if (configError != RTCErrorType.None)
            {
                throw new Exception("Unable to set WebRTC configuration");
            }

            // PLAT-1009
            // The WebRTC package supports receiving as well as sending video streams, but with different codecs (H264 for sending, VP8 for receiving).
            // The Hardware Encoder only supports H264 (which is what we want anyways). 
            // Forcing this into receive (from the browser pov)  forces the correct codecs.
            offer.sdp = offer.sdp.Replace("a=sendrecv", "a=recvonly");
            var setRemoteDescriptionOp = pc.SetRemoteDescription(ref offer);
            WaitForAsyncOperation(setRemoteDescriptionOp);

            foreach (var pair in m_listVideoStreamTrackUpdaters)
            {
                var sender = pc.AddTrack(pair.Key);
                m_rtpVideoSenders[pc] = sender;
            }

            if (m_audioStream != null)
            {
                foreach (var track in m_audioStream.GetTracks())
                {
                    pc.AddTrack(track);
                }
            }

            RTCAnswerOptions options = default;
            var createAnswserOp = pc.CreateAnswer(ref options);
            WaitForAsyncOperation(createAnswserOp);


            var desc = createAnswserOp.Desc;
            signaling.SendAnswer(connectionId, desc);

            var setLocationDescriptionOp = pc.SetLocalDescription(ref desc);
            WaitForAsyncOperation(setLocationDescriptionOp);
        }

        static void WaitForAsyncOperation(AsyncOperationBase op)
        {
            while (op.MoveNext())
            {
            }
            if (op.IsError)
            {
                throw new Exception(op.Error.ToString());
            }
        }

        void OnIceCandidate(Signaling signalling, PlatformIceCandidateDescription candidateDescription)
        {
            if (!m_mapConnectionIdAndPeer.TryGetValue(candidateDescription.playerId, out var pc))
            {
                Debug.Log("Received ICE candidate, but no active peer connection:" + candidateDescription.playerId);
                return;
            }

            RTCIceCandidate​ _candidate = new RTCIceCandidate(new RTCIceCandidateInit
            {
                candidate = candidateDescription.candidate.candidate,
                sdpMid = candidateDescription.candidate.sdpMid,
                sdpMLineIndex = candidateDescription.candidate.sdpMLineIndex
            });
            Debug.Log("Adding ICE Candidate: " + _candidate.ToString());
            pc.AddIceCandidate(_candidate);
        }

        void OnDataChannel(RTCPeerConnection pc, int playerId, RTCDataChannel channel)
        {
            Debug.Log("OnDataChannel");
            if (!m_mapPeerAndChannelDictionary.TryGetValue(pc, out var channels))
            {
                channels = new DataChannelDictionary();
                m_mapPeerAndChannelDictionary.Add(pc, channels);
                channels.Add(channel.Id, channel);
            }


            // Platform monitor connection
            if (playerId == 1)
            {
                Debug.Log("Establishing Platform Communications");
                channel.OnOpen = () =>
                {
                    Debug.Log("Platform data channel open");
                };
                channel.OnMessage = bytes => HandleAppServerMessage(channel, bytes);
                channel.OnClose = () =>
                {
                    Debug.Log("Platform data channel closed.");
                };
            }
            else
            {
                Debug.Log("Registering input handlers");
                PureWebInputAdapter input = PureWebInputAdapterFactory.Create();

                channel.OnMessage = bytes => input.ProcessInput(bytes, playerId, HandleInteractionEvent);
                channel.OnClose = () => OnCloseChannel(channel);


                try
                {
                    var obj = GameObject.FindGameObjectWithTag(PureWebTags.PureWebInputHandler);
                    if (obj != null)
                    {
                        var controller = obj.GetComponent<IPureWebInputHandler>();
                        if (controller != null)
                        {
                            controller.SetInput(input);
                        }
                    }
                }
                finally
                {
                    m_mapChannelAndRemoteInput[channel] = input;
                    PlayerConnectedEvent?.Invoke(this, new PlayerConnectedEventArgs() { PlayerId = playerId });
                }
            }
        }

        /// <summary>
        /// Broadcast structured data to all connected client
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="data"></param>
        public void Broadcast<T>(T data)
        {
            var json = JsonUtility.ToJson(data);
            var size = 1 + 2 * json.Length;
            var message = new byte[size];

            message[0] = 1; // Cirrus Response

            var jsonBytes = UnicodeEncoding.Unicode.GetBytes(json);
            Array.Copy(jsonBytes, 0, message, 1, jsonBytes.Length);

            foreach (var channel in m_mapChannelAndRemoteInput.Keys)
            {
                if (channel.ReadyState == RTCDataChannelState.Open)
                {
                    channel.Send(message);
                }
            }
        }

        void HandleInteractionEvent(InteractionEventArgs args)
        {
            if (InteractionEvent != null)
            {
                InteractionEvent(this, args);
            }
        }

        void HandleAppServerMessage(RTCDataChannel channel, byte[] message)
        {
            var request = PlatformMessaging.Deserialize<PlatformRequest>(message);
            switch (request.type)
            {
                case PlatformRequestTypes.RequestProcessInfo:
                    HandleProcessInfoRequest(channel);
                    break;
                case PlatformRequestTypes.RequestShutdown:
                    Debug.Log("RequestShutdown: Shutting down");
                    Application.Quit(0);
                    break;
                default:
                    Debug.LogFormat("Unsupported message: " + request.type);
                    break;
            }

        }

        void HandleProcessInfoRequest(RTCDataChannel channel)
        {
            var processInfo = new ProcessInfo
            {
                appId = System.Environment.GetEnvironmentVariable("PUREWEB_APPID"),
                appName = System.Environment.GetEnvironmentVariable("PUREWEB_APPNAME"),
                processId = System.Diagnostics.Process.GetCurrentProcess().Id,
                schedulerId = System.Environment.GetEnvironmentVariable("PUREWEB_SCHEDULERID"),
                serviceManagerId = System.Environment.GetEnvironmentVariable("PUREWEB_SERVICEMGRID")
            };

            if (string.IsNullOrEmpty(processInfo.appName))
            {
                processInfo.appName = "pluginbuilder";
            }

            if (string.IsNullOrEmpty(processInfo.appId))
            {
                processInfo.appId = Guid.NewGuid().ToString();
            }

            channel.Send(PlatformMessaging.Serialize(processInfo));
        }

        void OnCloseChannel(RTCDataChannel channel)
        {
            Debug.Log("Data channel close");
            if (m_mapChannelAndRemoteInput.TryGetValue(channel, out var input))
            {
                // Detach remote input
                PureWebInputAdapterFactory.Destroy(input);

                var controller = GetComponent<IPureWebInputHandler>();
                if (controller != null)
                {
                    controller.SetInput(null);
                }
            }
        }
    }
}