﻿using System;
using UnityEngine.InputSystem;
using UnityEngine.InputSystem.LowLevel;
using System.Collections.Generic;
using System.Linq;
using UnityEngine.InputSystem.Users;
using UnityEngine;
using UnityEngine.EventSystems;
using UnityEngine.UI;
using System.Reflection;
using UnityEngine.SceneManagement;

namespace PureWeb
{

    enum MessageType
    {
        /**********************************************************************/

        /*
         * Control Messages. Range = 0..49.
         */
        IFrameRequest = 0,
        RequestQualityControl = 1,
        MaxFpsRequest = 2,
        AverageBitrateRequest = 3,
        StartStreaming = 4,
        StopStreaming = 5,

        /**********************************************************************/

        /*
         * Input Messages. Range = 50..89.
         */

        // Generic Input Messages. Range = 50..59.
        UIInteraction = 50,

        // Keyboard Input Message. Range = 60..69.
        KeyDown = 60,
        KeyUp = 61,
        KeyPress = 62,

        // Mouse Input Messages. Range = 70..79.
        MouseEnter = 70,
        MouseLeave = 71,
        MouseDown = 72,
        MouseUp = 73,
        MouseMove = 74,
        MouseWheel = 75,

        // Touch Input Messages. Range = 80..89.
        TouchStart = 80,
        TouchEnd = 81,
        TouchMove = 82
    };


    struct InputFieldReference
    {
        public WeakReference<Component> Component;
        public MethodInfo ProcessEvent;
        public MethodInfo UpdateLabel;
    }
    /// <summary>
    /// Remote input adapter compatible with Pixel Streaming messaging.
    /// </summary>
    public class PureWebInputAdapter : IRemoteInput
    {
        // Track selection of uGUI elements and dispatch events directly to them.
        // In particular for InputFields / Keyboard events, which query local hardware.
        private GameObject m_selectedObject = new GameObject();
        private KeyboardState m_keys = new KeyboardState();
        private MouseState m_prevMouseState = new MouseState();
        private Dictionary<GameObject, InputFieldReference> m_keyboardTargetCache = new Dictionary<GameObject, InputFieldReference>();

        public Mouse RemoteMouse { get; }
        public Keyboard RemoteKeyboard { get; }
        public Touchscreen RemoteTouchscreen { get; }
        public Gamepad RemoteGamepad { get; }


        public PureWebInputAdapter(ref InputUser user)
        {
            RemoteMouse = user.pairedDevices.FirstOrDefault(device => device is Mouse) as Mouse;
            RemoteKeyboard = user.pairedDevices.FirstOrDefault(device => device is Keyboard) as Keyboard;
            RemoteTouchscreen = user.pairedDevices.FirstOrDefault(device => device is Touchscreen) as Touchscreen;
            RemoteGamepad = user.pairedDevices.FirstOrDefault(device => device is Gamepad) as Gamepad;

            SceneManager.activeSceneChanged += (s,e) => {
                m_keyboardTargetCache = new Dictionary<GameObject, InputFieldReference>();
            };
        }

        

        // See:
        // https://developer.mozilla.org/en-US/docs/Web/API/MouseEvent/button
        private static Dictionary<byte, MouseButton> MouseButtonMap = new Dictionary<byte, MouseButton>() {
            {0, MouseButton.Left},
            {1, MouseButton.Middle },
            {2, MouseButton.Right },
            {3, MouseButton.Forward },
            {4, MouseButton.Back },
        };
        private void ProcessMouseEvent(byte[] bytes)
        {
            MouseButton button;
            ushort x, y;
            short deltaX, deltaY;
            Vector2Int pos;
            Vector2Int delta;
            MouseState newMouseState;

            switch ((MessageType)bytes[0])
            {
                case MessageType.MouseDown:
                    button = MouseButtonMap[bytes[1]];
                    x = (ushort)((BitConverter.ToUInt16(bytes, 2) / 65536.0) * Screen.width);
                    y = (ushort)(Screen.height - (BitConverter.ToUInt16(bytes, 4) / 65536.0) * Screen.height);

                    pos = new Vector2Int(x, y);
                    newMouseState = new MouseState
                    {
                        position = pos,
                        delta = pos - m_prevMouseState.position,
                        buttons = m_prevMouseState.buttons
                    };

                    newMouseState.WithButton(button, true);

                    InputSystem.QueueStateEvent(RemoteMouse, newMouseState);
                    m_prevMouseState = newMouseState;

                    break;
                case MessageType.MouseUp:
                    button = MouseButtonMap[bytes[1]];
                    x = (ushort)((BitConverter.ToUInt16(bytes, 2) / 65536.0) * Screen.width);
                    y = (ushort)(Screen.height - (BitConverter.ToUInt16(bytes, 4) / 65536.0) * Screen.height);

                    pos = new Vector2Int(x, y);
                    newMouseState = new MouseState
                    {
                        position = m_prevMouseState.position,
                        delta = pos - m_prevMouseState.position,
                        buttons = m_prevMouseState.buttons
                    };

                    newMouseState.WithButton(button, false);

                    InputSystem.QueueStateEvent(RemoteMouse, newMouseState);
                    m_prevMouseState = newMouseState;

                    break;
                case MessageType.MouseMove:

                    // Values come from Unreal Noramlized to [0,65536]
                    // 0,0 in Unity is bottom left
                    // 0,0 in Unreal is top left
                    // This means that deltaY needs to be negated in Unity
                    x = (ushort)((BitConverter.ToUInt16(bytes, 1) / 65536.0) * Screen.width);
                    y = (ushort)(Screen.height - (BitConverter.ToUInt16(bytes, 3) / 65536.0) * Screen.height);
                    deltaX = (short)((BitConverter.ToInt16(bytes, 5) / 32767.0) * Screen.width);
                    deltaY = (short)((BitConverter.ToInt16(bytes, 7) / 32767.0) * Screen.height * -1);

                    pos = new Vector2Int(x, y);
                    delta = new Vector2Int(deltaX, deltaY);

                    newMouseState = new MouseState
                    {
                        position = pos,
                        delta = delta,
                        buttons = m_prevMouseState.buttons
                    };
                    InputSystem.QueueStateEvent(RemoteMouse, newMouseState);
                    m_prevMouseState = newMouseState;
                    break;
                default:
                    break;
            }

        }

        public static readonly Dictionary<int, Key> KeyMap = new Dictionary<int, Key>() {
           {0, Key.None },
           {32, Key.Space },
           {13, Key.Enter },
           {9, Key.Tab },
           {192, Key.Backquote },
           {222, Key.Quote },
           {186, Key.Semicolon },
           {188, Key.Comma},
           {190, Key.Period },
           {191, Key.Slash },
           {220, Key.Backslash },
           {219, Key.LeftBracket },
           {221, Key.RightBracket },
           {189, Key.Minus },
           {187, Key.Equals },
           {65, Key.A },
           {66, Key.B },
           {67, Key.C },
           {68, Key.D },
           {69, Key.E },
           {70, Key.F },
           {71, Key.G },
           {72, Key.H },
           {73, Key.I },
           {74, Key.J },
           {75, Key.K },
           {76, Key.L },
           {77, Key.M },
           {78, Key.N },
           {79, Key.O },
           {80, Key.P },
           {81, Key.Q },
           {82, Key.R },
           {83, Key.S },
           {84, Key.T },
           {85, Key.U },
           {86, Key.V },
           {87, Key.W },
           {88, Key.X },
           {89, Key.Y },
           {90, Key.Z },
           {49, Key.Digit1 },
           {50, Key.Digit2 },
           {51, Key.Digit3 },
           {52, Key.Digit4 },
           {53, Key.Digit5 },
           {54, Key.Digit6 },
           {55, Key.Digit7 },
           {56, Key.Digit8 },
           {57, Key.Digit9 },
           {48, Key.Digit0 },
           {16, Key.LeftShift },
           //{16, Key.RightShift },
           {18, Key.LeftAlt },
           //{18, Key.RightAlt },
           //{0, Key.AltGr },
           {17, Key.LeftCtrl },
           //{17, Key.RightCtrl },
           {91, Key.LeftMeta },
           //{91, Key.LeftWindows },
           //{91, Key.LeftApple },
           //{91, Key.LeftCommand },
           {92, Key.RightMeta },
           //{92, Key.RightWindows },
           //{92, Key.RightApple },
           //{92, Key.RightCommand },
           {93, Key.ContextMenu },
           {27, Key.Escape },
           {37, Key.LeftArrow },
           {39, Key.RightArrow },
           {38, Key.UpArrow },
           {40, Key.DownArrow },
           {8, Key.Backspace },
           {34, Key.PageDown },
           {33, Key.PageUp },
           {36, Key.Home },
           {35, Key.End },
           {45, Key.Insert },
           {46, Key.Delete },
           {20, Key.CapsLock },
           {144, Key.NumLock },
           {44, Key.PrintScreen },
           {145, Key.ScrollLock },
           {19, Key.Pause },
           //{13, Key.NumpadEnter },
           {111, Key.NumpadDivide },
           {106, Key.NumpadMultiply },
           {107, Key.NumpadPlus },
           {109, Key.NumpadMinus },
           {110, Key.NumpadPeriod },
           //{0, Key.NumpadEquals },
           {96, Key.Numpad0 },
           {97, Key.Numpad1 },
           {98, Key.Numpad2 },
           {99, Key.Numpad3 },
           {100, Key.Numpad4 },
           {101, Key.Numpad5 },
           {102, Key.Numpad6 },
           {103, Key.Numpad7 },
           {104, Key.Numpad8 },
           {105, Key.Numpad9 },
           {112, Key.F1 },
           {113, Key.F2 },
           {114, Key.F3 },
           {115, Key.F4 },
           {116, Key.F5 },
           {117, Key.F6 },
           {118, Key.F7 },
           {119, Key.F8 },
           {120, Key.F9 },
           {121, Key.F10 },
           {122, Key.F11 },
           {123, Key.F12 },
           {146, Key.OEM1 },
           {147, Key.OEM2 },
           {148, Key.OEM3 },
           {149, Key.OEM4 },
           {150, Key.OEM5 },
           //{0, Key.IMESelected}
        };

        public static readonly Dictionary<int, KeyCode> KeyCodeMap = new Dictionary<int, KeyCode>() {
           {0, KeyCode.None },
           {32, KeyCode.Space },
           {13, KeyCode.Return },
           {9, KeyCode.Tab },
           {192, KeyCode.BackQuote },
           {222, KeyCode.Quote },
           {186, KeyCode.Semicolon },
           {188, KeyCode.Comma},
           {190, KeyCode.Period },
           {191, KeyCode.Slash },
           {220, KeyCode.Backslash },
           {219, KeyCode.LeftBracket },
           {221, KeyCode.RightBracket },
           {189, KeyCode.Minus },
           {187, KeyCode.Equals },
           {65, KeyCode.A },
           {66, KeyCode.B },
           {67, KeyCode.C },
           {68, KeyCode.D },
           {69, KeyCode.E },
           {70, KeyCode.F },
           {71, KeyCode.G },
           {72, KeyCode.H },
           {73, KeyCode.I },
           {74, KeyCode.J },
           {75, KeyCode.K },
           {76, KeyCode.L },
           {77, KeyCode.M },
           {78, KeyCode.N },
           {79, KeyCode.O },
           {80, KeyCode.P },
           {81, KeyCode.Q },
           {82, KeyCode.R },
           {83, KeyCode.S },
           {84, KeyCode.T },
           {85, KeyCode.U },
           {86, KeyCode.V },
           {87, KeyCode.W },
           {88, KeyCode.X },
           {89, KeyCode.Y },
           {90, KeyCode.Z },
           {49, KeyCode.Alpha1 },
           {50, KeyCode.Alpha2 },
           {51, KeyCode.Alpha3 },
           {52, KeyCode.Alpha4 },
           {53, KeyCode.Alpha5 },
           {54, KeyCode.Alpha6 },
           {55, KeyCode.Alpha7 },
           {56, KeyCode.Alpha8 },
           {57, KeyCode.Alpha9 },
           {48, KeyCode.Alpha0 },
           {16, KeyCode.LeftShift },
           //{16, KeyCode.RightShift },
           {18, KeyCode.LeftAlt },
           //{18, KeyCode.RightAlt },
           //{0, KeyCode.AltGr },
           {17, KeyCode.LeftControl },
           //{17, KeyCode.RightCtrl },
           {91, KeyCode.LeftWindows},
           //{91, KeyCode.LeftWindows },
           //{91, KeyCode.LeftApple },
           //{91, KeyCode.LeftCommand },
           {92, KeyCode.RightWindows },
           //{92, KeyCode.RightWindows },
           //{92, KeyCode.RightApple },
           //{92, KeyCode.RightCommand },
           {93, KeyCode.Menu },
           {27, KeyCode.Escape },
           {37, KeyCode.LeftArrow },
           {39, KeyCode.RightArrow },
           {38, KeyCode.UpArrow },
           {40, KeyCode.DownArrow },
           {8, KeyCode.Backspace },
           {34, KeyCode.PageDown },
           {33, KeyCode.PageUp },
           {36, KeyCode.Home },
           {35, KeyCode.End },
           {45, KeyCode.Insert },
           {46, KeyCode.Delete },
           {20, KeyCode.CapsLock },
           {144, KeyCode.Numlock },
           {44, KeyCode.Print },
           {145, KeyCode.ScrollLock },
           {19, KeyCode.Pause },
           {12, KeyCode.Clear },
           //{13, KeyCode.KeypadEnter },
           {111, KeyCode.KeypadDivide },
           {106, KeyCode.KeypadMultiply },
           {107, KeyCode.KeypadPlus },
           {109, KeyCode.KeypadMinus },
           {110, KeyCode.KeypadPeriod },
           //{0, KeyCode.KeypadEquals },
           {96, KeyCode.Keypad0 },
           {97, KeyCode.Keypad1 },
           {98, KeyCode.Keypad2 },
           {99, KeyCode.Keypad3 },
           {100, KeyCode.Keypad4 },
           {101, KeyCode.Keypad5 },
           {102, KeyCode.Keypad6 },
           {103, KeyCode.Keypad7 },
           {104, KeyCode.Keypad8 },
           {105, KeyCode.Keypad9 },
           {112, KeyCode.F1 },
           {113, KeyCode.F2 },
           {114, KeyCode.F3 },
           {115, KeyCode.F4 },
           {116, KeyCode.F5 },
           {117, KeyCode.F6 },
           {118, KeyCode.F7 },
           {119, KeyCode.F8 },
           {120, KeyCode.F9 },
           {121, KeyCode.F10 },
           {122, KeyCode.F11 },
           {123, KeyCode.F12 },
           //{146, KeyCode.OEM1 },
           //{147, KeyCode.OEM2 },
           //{148, KeyCode.OEM3 },
           //{149, KeyCode.OEM4 },
           //{150, KeyCode.OEM5 },
           //{0, Key.IMESelected}
        };

        

        private void ProcessKeyEvent(byte[] bytes)
        {
            // Update keyboard state
            var messageType = (MessageType)bytes[0];
            var javascriptKeyCode = bytes[1];
            switch ((MessageType)bytes[0])
            {
                case MessageType.KeyDown:

                    var repeat = bytes[2] != 0;
                    if (!repeat)
                    {

                        if (KeyMap.TryGetValue(javascriptKeyCode, out var keyDown))
                        {
                            m_keys.Set(keyDown, true);
                            InputSystem.QueueStateEvent(RemoteKeyboard, m_keys);
                        }
                    }
                    break;
                case MessageType.KeyPress:
                    var characterCode = (char)BitConverter.ToUInt16(bytes, 1);
                    InputSystem.QueueTextEvent(RemoteKeyboard, characterCode);
                    break;
                case MessageType.KeyUp:
                    if (KeyMap.TryGetValue(javascriptKeyCode, out var keyUp))
                    {
                        m_keys.Set(keyUp, false);
                        InputSystem.QueueStateEvent(RemoteKeyboard, m_keys);
                    }
                    break;
            }

            
            GameObject selectedObject = EventSystem.current.currentSelectedGameObject;
            if (selectedObject == null) return;

            // Direct message directly to the input field, if applicable

            MethodInfo processEvent = null;
            MethodInfo updateLabel = null;
            Component target = null;

            if (m_keyboardTargetCache.ContainsKey(selectedObject))
            {
                var reference = m_keyboardTargetCache[selectedObject];
                reference.Component.TryGetTarget(out target);
                

                if (target == null)
                {
                    // Purge cache
                    m_keyboardTargetCache.Remove(selectedObject);
                }
                else
                {
                    processEvent = reference.ProcessEvent;
                    updateLabel = reference.UpdateLabel;
                }
            }

            if (target == null)
            {
                foreach (var c in selectedObject.GetComponents<Component>())
                {
                    processEvent = c.GetType().GetMethod("ProcessEvent");
                    updateLabel = c.GetType().GetMethod("ForceLabelUpdate");
                    target = c;

                    // Found an InputField
                    if (processEvent != null && updateLabel != null)
                    {
                        m_keyboardTargetCache[selectedObject] = new InputFieldReference() {
                            Component = new WeakReference<Component>(target),
                            ProcessEvent = processEvent,
                            UpdateLabel = updateLabel
                        };
                    }
                }
            }
           

            if (processEvent == null || updateLabel == null)
            {
                // Not a InputField-like selected object
                return;
            }

            switch (messageType)
            {
                case MessageType.KeyDown:
                    if (!KeyCodeMap.TryGetValue(bytes[1], out var keyCode))
                    {
                        keyCode = KeyCode.None;
                        Debug.LogFormat("Unhandled key {0}", bytes[1]);
                    }
                    var keyDownEvent = new Event
                    {
                        alt = RemoteKeyboard.altKey.isPressed,
                        control = RemoteKeyboard.ctrlKey.isPressed,
                        shift = RemoteKeyboard.shiftKey.isPressed,
                        command = RemoteKeyboard.leftMetaKey.isPressed || RemoteKeyboard.rightMetaKey.isPressed,
                        character = default,
                        keyCode = keyCode
                    };

                    processEvent.Invoke(target, new object[] { keyDownEvent });
                    break;
                case MessageType.KeyPress:
                    var characterCode = (char)BitConverter.ToUInt16(bytes, 1);
                    var keyPressEvent = new Event
                    {
                        alt = RemoteKeyboard.altKey.isPressed,
                        control = RemoteKeyboard.ctrlKey.isPressed,
                        shift = RemoteKeyboard.shiftKey.isPressed,
                        command = RemoteKeyboard.leftMetaKey.isPressed || RemoteKeyboard.rightMetaKey.isPressed,
                        character = characterCode,
                        keyCode = characterCode == 8 ? KeyCode.Backspace : KeyCode.None // Special handling for backspace because Unity UI wrongly examines the keycode
                    };

                    processEvent.Invoke(target, new object[] { keyPressEvent });
                    break;
            }

            updateLabel.Invoke(target, null);
        }
        public void ProcessInput(byte[] bytes, int playerId, Action<InteractionEventArgs> ProcessInteractionEvent)
        {
            switch ((MessageType)bytes[0])
            {
                case MessageType.MouseEnter:
                case MessageType.MouseLeave:
                case MessageType.MouseDown:
                case MessageType.MouseUp:
                case MessageType.MouseWheel:
                case MessageType.MouseMove:
                    ProcessMouseEvent(bytes);
                    break;
                case MessageType.KeyDown:
                case MessageType.KeyUp:
                case MessageType.KeyPress:
                    ProcessKeyEvent(bytes);
                    break;
                case MessageType.TouchStart:
                case MessageType.TouchMove:
                case MessageType.TouchEnd:
                    // TODO: TouchEvents
                    break;
                case MessageType.UIInteraction:
                    InteractionEventArgs args = new InteractionEventArgs(playerId, bytes);

                    ProcessInteractionEvent(args);
                    break;
                default:
                    Debug.LogFormat("Unhandled input message type {0}", bytes[0]);
                    break;
            }

        }
    }
}
